# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""Usage: unique [-c]

Copies input objects to output, dropping duplicates. No output is
generated until the end of the input stream occurs. However, if
the duplicates are known to be consecutive, then specifying -c
allows output to be generated sooner.
"""

import types

import osh.oshargs
import osh.oshobjects

def unique(*args):
    return _Unique().process_args(*args)

class _Unique(osh.oshobjects.Op):

    _uniquer = None


    # object interface

    def __init__(self):
        osh.oshobjects.Op.__init__(self, 'c', (0, 0))


    # BaseOp interface

    def doc(self):
        return __doc__

    def setup(self):
        args = self.args()
        if args.has_next():
            self.usage()
        if args.flag('-c'):
            self._uniquer = _ConsecutiveUniquer(self)
        else:
            self._uniquer = _GeneralUniquer(self)

    def receive(self, object):
        self._uniquer.receive(object)

    def receive_complete(self):
        self._uniquer.receive_complete()


class _GeneralUniquer:

    _command = None
    _unique = None

    def __init__(self, command):
        self._command = command
        self._unique = {}

    def receive(self, object):
        if type(object) == types.ListType:
            object = tuple(object)
        self._unique[object] = None

    def receive_complete(self):
        for key in self._unique.keys():
            self._command.send(key)

class _ConsecutiveUniquer:

    _command = None
    _current = None

    def __init__(self, command):
        self._command = command

    def receive(self, object):
        if self._current != object:
            if self._current != None:
                self._command.send(self._current)
            self._current = object

    def receive_complete(self):
        if self._current != None:
            self._command.send(self._current)
