# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""osh function syntax:

    [[lambda] ARGS:] EXPRESSION

ARGS: Comma-separated list of function arguments, optionally preceded by keyword
'lambda'. ARGS may be omitted only for a function with no arguments.

EXPRESSION: A Python expression written in terms of the function arguments.

Example:

    x: max(x, 10)
"""

import copy
import inspect
import sys
import types

import oshobjects

# Populated by the imp command
defined_by_osh = {}

LAMBDA = 'lambda '
LAMBDA_COLON = 'lambda:'
_function_counter = 0

class OshFunction:
    """Represents a function for use by osh commands with function arguments,
    e.g., select.
    """
    
    _function = None
    _function_id = None
    _no_args = None

    def __init__(self, function_spec, namespace = oshobjects.__dict__):
        """Creates an OshFunction.
        - function_spec: String specification of a function. Uses lambda notation,
          but the 'lambda' keyword is optional.
        - namespace: The default is reasonable; there's been no need to
          override it so far.
        """
        global _function_counter
        if type(function_spec) is types.FunctionType:
            self._function = function_spec
        else:
            lambda_expression = self.parse(function_spec.strip())
            # Assign a function id
            self._function_id = _function_counter
            _function_counter += 1
            # Create an n(initialValue = 0) function and place it in the
            # OshFunction's namespace.
            namespace = copy.copy(namespace)
            namespace['n'] = (lambda initialValue = 0:
                              oshobjects.n(self._function_id, initialValue))
            namespace.update(defined_by_osh)
            self._function = eval(lambda_expression, namespace)
        argspec = inspect.getargspec(self._function)
        self._no_args = len(argspec[0]) == 0 and argspec[1] is None and argspec[2] is None

    def __call__(self, args = None):
        if self._no_args:
            return self._function()
        else:
            return self._function(*args)

    def parse(self, function_spec):
        # If the function spec starts with a lambda, then just use it as is.
        # Otherwise, it could be:
        # - function with arg list, e.g. "x, y: x + y"
        # - zero-arg list with colon, e.g. ": processes()"
        # - zero-arg list without colon, e.g. "processes()"
        # Try prepending "lambda:" and "lambda" until there's
        # no syntax error. Crude but effective.
        if function_spec.startswith(LAMBDA) or function_spec.startswith(LAMBDA_COLON):
            # ok as is
            pass
        else:
            fixed_function_spec = '%s: %s' % (LAMBDA, function_spec)
            try:
                eval(fixed_function_spec)
                function_spec = fixed_function_spec
            except:
                fixed_function_spec = '%s %s' % (LAMBDA, function_spec)
                try:
                    eval(fixed_function_spec)
                    function_spec = fixed_function_spec
                except:
                    raise Exception('Illegal function spec: %s' % function_spec)
        return function_spec
