# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import cPickle
import os
import popen2
import sys
import threading
import traceback

# OshProcess coordinates with consumer threads (processing process stdout
# and stderr) through the use of a condition var. It would be simpler to
# just join the consumer threads, but this seems not to work (python2.2
# on rh9), sometimes getting stuck in join, requiring the process be killed.

all_processes = []

def kill_all_processes():
    # print '>>> Killing %s processes' % len(all_processes)
    count = 0
    for process in all_processes:
        count += 1
        # print '>>> Killing #%s' % count
        try:
            process.kill()
        except:
            traceback.print_exc(file = sys.stderr)

class OshProcess:

    _command = None
    _process = None
    _input_provider = None
    _out_consumer = None
    _err_consumer = None
    _process_completion = None
    _terminating_exception = None

    def __init__(self,
                 command,
                 input_provider,
                 out_consumer,
                 err_consumer):
        self._command = command
        self._input_provider = input_provider
        self._out_consumer = out_consumer
        self._err_consumer = err_consumer
        self._process_completion = threading.Condition(threading.RLock())

    def dump(self, label):
        print '%s: %s' % (label, self._command)

    def pid(self):
        return self._process.pid

    def run(self):        
        try:
            try:
                self._process = popen2.Popen3(self._command, True)
                all_processes.append(self)
                # print '>>> Starting process. Running: %s' % len(all_processes)
                # send input to process
                self._input_provider.initialize(self._process.tochild, self)
                self._input_provider.run()
                # Flush but don't close the stream -- may need to send
                # kill signal later.
                self._process.tochild.flush()
                # set up out and err consumers
                self._out_consumer.initialize(self._process.fromchild, self)
                self._err_consumer.initialize(self._process.childerr, self)
                self._out_consumer.start()
                self._err_consumer.start()
            except Exception, e:
                # print '>>> Process ended with exception %s. Running: %s' % (e, len(all_processes))
                self._terminating_exception = e
        finally:
            # wait for the process and consumer threads to end
            try:
                if self._process:
                    exitCode = self._process.wait()
                    self._wait_for_consumers_to_finish()
                    self.close_input_stream()
                    self._out_consumer.join()
                    self._err_consumer.join()
                # print '>>> Process ended normally. Running: %s' % len(all_processes)
            except OSError, e:
                # Process can disappear before we check exit code
                # print '>>> OSError during wait for process end, %s. Running: %s' % (e, len(all_processes))
                pass
            all_processes.remove(self)

    def close_input_stream(self):
        self._process.tochild.flush()
        self._process.tochild.close()


    def kill(self):
        self._input_provider.send_kill(9)
        os.kill(self._process.pid, 9)

    def terminating_exception(self):
        return self._terminating_exception

    # for use by this class

    def _wait_for_consumers_to_finish(self):
        try:
            self._process_completion.acquire()
            while not (self._out_consumer.done() and self._err_consumer.done()):
                self._process_completion.wait(1.0)
        finally:
            self._process_completion.release()

class _StreamHandler:

    _handler = None
    _process = None
    _stream = None
    _done = None

    def __init__(self, handler):
        self._handler = handler
        self._done = False

    def initialize(self, process):
        self._process = process

    def stream(self):
        return self._stream;
 
    def handler(self):
        return self._handler

    def command(self):
        return self._process._command

    def terminating_exception(self, exception):
        self._process._terminating_exception = exception

    def done(self):
        return self._done

    def notify_process_of_completion(self):
        self._process._process_completion.acquire()
        self._done = True
        self._process._process_completion.notify()
        self._process._process_completion.release()

class ObjectInputProvider(_StreamHandler):

    _inputs = None

    def __init__(self, handler, inputs):
        _StreamHandler.__init__(self, handler)
        self._inputs = inputs

    def initialize(self, stream, process):
        _StreamHandler.initialize(self, process)
        self._stream = cPickle.Pickler(stream)

    def run(self):
        try:
            for input in self._inputs:
                self.handler()(self.stream(), input)
        except Exception, e:
            self.terminating_exception(e)

    def send_kill(self, kill_signal):
        self.handler()(self.stream(), kill_signal)
        self._process.close_input_stream()

class ObjectOutputConsumer(_StreamHandler, threading.Thread):

    _baseStream = None

    def __init__(self, handler):
        _StreamHandler.__init__(self, handler)
        threading.Thread.__init__(self)

    def initialize(self, stream, process):
        _StreamHandler.initialize(self, process)
        self._baseStream = stream
        self._stream = cPickle.Unpickler(stream)
        
    def run(self):
        try:
            try:
                eof = False
                while not eof:
                    try:
                        object = self.stream().load()
                        self.handler()(object)
                    except EOFError, e:
                        eof = True
            except Exception, e:
                self.terminating_exception(e)
        finally:
            self._baseStream.close()
            self.notify_process_of_completion()
        
class LineOutputConsumer(_StreamHandler, threading.Thread):

    def __init__(self, handler):
        _StreamHandler.__init__(self, handler)
        threading.Thread.__init__(self)

    def initialize(self, stream, process):
        _StreamHandler.initialize(self, process)
        self._stream = stream

    def run(self):
        try:
            try:
                eof = False
                while not eof:
                    line = self.stream().readline()
                    if line:
                        self.handler()(line)
                    else:
                        eof = True
            except Exception, e:
                self.terminating_exception(e)
        finally:
            self.stream().close()
            self.notify_process_of_completion()
